import os
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd

from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.inversion.results.utils.abstract_util import ResultsUtil
from beetroots.inversion.results.utils.perf_saver import EstimatorPerfSaver
from beetroots.modelling.posterior import Posterior
from beetroots.space_transform.abstract_transform import Scaler


class ResultsMMSEandCI(ResultsUtil):

    __slots__ = (
        "model_name",
        "path_img",
        "path_data_csv_out_mcmc",
        "N",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        path_img: str,
        path_data_csv_out_mcmc: str,
        N: int,
        D: int,
    ):
        self.model_name = model_name
        self.path_img = path_img
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc

        self.N = N
        self.D = D

    def read_data(self) -> pd.DataFrame:
        path_overall_results = f"{self.path_data_csv_out_mcmc}/"
        path_overall_results += f"estimation_Theta_{self.model_name}.csv"

        df_estim = pd.read_csv(path_overall_results, index_col=["n", "d"])
        df_estim = df_estim.sort_index().reset_index(drop=False)

        assert len(df_estim) == self.N * self.D
        return df_estim

    def create_folders(self, list_CI: List[int]) -> Tuple[str, dict[int, str]]:
        folder_path = f"{self.path_img}/estimators"
        folder_path_inter = f"{folder_path}/{self.model_name}"
        folder_path_MMSE = f"{folder_path_inter}/MMSE"

        dict_folder_path_CI = {ci: f"{folder_path_inter}/CI{ci}" for ci in list_CI}

        for path_ in [
            folder_path,
            folder_path_inter,
            folder_path_MMSE,
        ] + list(dict_folder_path_CI.values()):
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path_MMSE, dict_folder_path_CI

    def main(
        self,
        posterior: Posterior,
        scaler: Scaler,
        estimator_plot: Optional[PlotsEstimator],
        Theta_true_scaled_full: Optional[np.ndarray],
        list_idx_sampling: List[int],
        list_fixed_values: np.ndarray,
        list_CI: List[int],
    ) -> pd.DataFrame:
        estimator_name = "MMSE"

        df_estim = self.read_data()
        Theta_mmse_lin = df_estim["Theta_MMSE"].values.reshape((self.N, self.D))
        Theta_mmse_scaled_full = scaler.from_lin_to_scaled(Theta_mmse_lin)

        # save its performance
        perf_saver = EstimatorPerfSaver()
        if Theta_true_scaled_full is not None:
            mse = perf_saver.compute_MSE(
                Theta_mmse_scaled_full,
                Theta_true_scaled_full,
            )
            snr = perf_saver.compute_SNR(
                Theta_mmse_scaled_full,
                Theta_true_scaled_full,
            )
        else:
            mse = None
            snr = None

        # evaluate objective
        Theta_mmse_scaled = Theta_mmse_scaled_full[:, list_idx_sampling]

        forward_map_evals = posterior.likelihood.evaluate_all_forward_map(
            Theta_mmse_scaled,
            compute_derivatives=False,
            compute_derivatives_2nd_order=False,
        )
        nll_utils = posterior.likelihood.evaluate_all_nll_utils(
            forward_map_evals,
            compute_derivatives=False,
            compute_derivatives_2nd_order=False,
        )
        objective_mmse = posterior.neglog_pdf(
            Theta_mmse_scaled, forward_map_evals, nll_utils
        )

        perf_saver.save_estimator_performance(
            self.path_data_csv_out_mcmc,
            estimator_name,
            self.model_name,
            mse,
            snr,
            objective_mmse,
        )

        # plot it
        N, _ = Theta_mmse_lin.shape
        if N > 1 and estimator_plot is not None:
            (
                folder_path_MMSE,
                dict_folder_path_CI,
            ) = self.create_folders(list_CI)

            estimator_plot.plot_estimator(
                Theta_mmse_lin,
                estimator_name,
                folder_path_MMSE,
                self.model_name,
            )

            dict_ci_per = {
                ci: {"lower": (100 - ci) / 2, "upper": 100 - (100 - ci) / 2}
                for ci in list_CI
            }
            dict_per_ci = {}
            for ci in list_CI:
                per_lower = dict_ci_per[ci]["lower"] * 1
                per_upper = dict_ci_per[ci]["upper"] * 1
                dict_per_ci[per_lower] = ci * 1
                dict_per_ci[per_upper] = ci * 1

            list_per = list(dict_per_ci.keys())
            list_per.sort()

            for ci in list_CI:
                for percentile in dict_ci_per[ci].values():
                    estimator_name = f"percentile {percentile:.1f}%"
                    per_name = f"per_{percentile:.1f}".replace(".", "p")

                    Theta_percentile_lin = df_estim[per_name].values
                    Theta_percentile_lin = Theta_percentile_lin.reshape(
                        (self.N, self.D),
                    )

                    estimator_plot.plot_estimator(
                        Theta_percentile_lin,
                        estimator_name,
                        dict_folder_path_CI[ci],
                        self.model_name,
                    )

                per_low = dict_ci_per[ci]["lower"] * 1
                per_low_name = f"per_{per_low:.1f}".replace(".", "p")

                per_up = dict_ci_per[ci]["upper"] * 1
                per_up_name = f"per_{per_up:.1f}".replace(".", "p")

                df_ratio = df_estim[per_up_name] / df_estim[per_low_name]
                ratio = df_ratio.values.reshape((self.N, self.D))
                estimator_plot.plot_CI_size(
                    ratio,
                    CI_name=f"{ci}% CI size",
                    folder_path=dict_folder_path_CI[ci],
                )

                estimator_plot.plot_CI_size(
                    np.sqrt(ratio),
                    CI_name=f"{ci}% CI uncertainty factor",
                    folder_path=dict_folder_path_CI[ci],
                )

            # self.plot_map_nll_of_estimator(
            #     Theta_MAP_scaled, estimator_name, model_name
            # )
        return df_estim



"""Defines the Total variation spatial regularization and its derivatives
"""
import numba
import numpy as np
import pandas as pd

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


@numba.jit(nopython=True)  # , cache=True)
def compute_tv_matrix(
    Theta: np.ndarray, list_edges: np.ndarray, eps: float
) -> np.ndarray:
    tv_matrix_ = np.zeros_like(Theta) + eps

    N = tv_matrix_.shape[0]
    for i in range(N):
        # mask_i_m = list_edges[:, 1] == i
        mask_i_p = list_edges[:, 0] == i

        tv_matrix_[i] += np.sum(
            (Theta[list_edges[mask_i_p, 1], :] - Theta[list_edges[mask_i_p, 0], :])
            ** 2,
            axis=0,
        )
    return np.sqrt(tv_matrix_)  # (N, D)


@numba.jit(nopython=True)
def compute_gradient_from_tv_matrix(
    Theta: np.ndarray, tv_matrix_: np.ndarray, list_edges: np.ndarray
) -> np.ndarray:
    grad_ = np.zeros_like(Theta, dtype=np.float64)
    for edge in list_edges:
        val = (Theta[edge[1]] - Theta[edge[0]]) / tv_matrix_[edge[0]]  # (D,)
        grad_[edge[0], :] -= val
        grad_[edge[1], :] += val

    return grad_


class TVeps2DSpatialPrior(SpatialPrior):
    def __init__(
        self,
        D: int,
        N: int,
        df: pd.DataFrame,
        weights: np.ndarray = None,
        eps: float = 1e-3,
    ):
        super().__init__(D, N, df, weights)
        self.eps = eps

    def neglog_pdf(self, Theta: np.ndarray, with_weights: bool = True) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        if self.list_edges.size == 0:
            return np.zeros((self.D,))

        tv_matrix_ = compute_tv_matrix(Theta, self.list_edges, self.eps)
        nlpdf = np.sum(tv_matrix_, axis=0)

        if with_weights:
            nlpdf *= self.weights

        return nlpdf  # (D,)

    def gradient_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)
        if self.list_edges.size == 0:
            return np.zeros_like(Theta, dtype=np.float64)

        tv_matrix_ = compute_tv_matrix(Theta, self.list_edges, self.eps)  # (N, D)

        grad_ = compute_gradient_from_tv_matrix(Theta, tv_matrix_, self.list_edges)
        grad_ = grad_ * self.weights[None, :]
        return grad_  # (N, D)

    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        # TODO
        assert Theta.shape == (self.N, self.D)

        hess_diag = np.zeros_like(Theta, dtype=np.float64)

        #! unfinished
        # for edge in self.list_edges:
        #     # val = 1 / np.sqrt((x[edge[1]] - x[edge[0]]) ** 2 + self.eps) - (
        #     #     x[edge[1]] - x[edge[0]]
        #     # ) ** 2 * ((x[edge[1]] - x[edge[0]]) ** 2 + self.eps) ** (
        #     #     -3 / 2
        #     # )  # (D,)
        #     delta_x = x[edge[1]] - x[edge[0]]
        #     val = 2 * self.eps / (delta_x ** 2 + self.eps) ** (3 / 2)
        #     hess_diag[edge[0], :] += val
        #     hess_diag[edge[1], :] += val

        hess_diag = hess_diag * self.weights[None, :]
        return hess_diag



import os
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.inversion.results.abstract_results import ResultsExtractor
from beetroots.inversion.results.utils.bayes_pval_plots import ResultsBayesPvalues
from beetroots.inversion.results.utils.clppd import ResultsCLPPD
from beetroots.inversion.results.utils.kernel import ResultsKernels
from beetroots.inversion.results.utils.lowest_obj_estimator import (
    ResultsLowestObjective,
)
from beetroots.inversion.results.utils.objective import ResultsObjective
from beetroots.modelling.posterior import Posterior
from beetroots.space_transform.abstract_transform import Scaler


class ResultsExtractorOptimMAP(ResultsExtractor):
    r"""extractor of inference results for the data of the optimization runs that that was saved."""
    __slots__ = (
        "path_data_csv_out_optim_map",
        "path_img",
        "path_raw",
        "N_MCMC",
        "T_MC",
        "T_BI",
        "freq_save",
        "max_workers",
    )

    def __init__(
        self,
        path_data_csv_out_optim_map: str,
        path_img: str,
        path_raw: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        max_workers: int,
    ):
        r"""

        Parameters
        ----------
        path_data_csv_out_optim_map : str
            path to the csv file in which the performance of estimators is to be saved
        path_img : str
            path to the folder in which images are to be saved
        path_raw : str
            path to the raw ``.hdf5`` files
        N_MCMC : int
            number of optimization procedures to run per posterior distribution
        T_MC : int
            total size of each optimization procedure
        T_BI : int
            duration of the `Burn-in` phase
        freq_save : int
            frequency of saved iterates, 1 means that all iterates were saved (used to show correct optimization procedure sizes in chain plots)
        max_workers : int
            maximum number of workers that can be used for results extraction
        """
        self.path_data_csv_out_optim_map = path_data_csv_out_optim_map
        r"""str: path to the csv file in which the performance of estimators is to be saved"""
        self.path_img = path_img
        r"""str: path to the folder in which images are to be saved"""
        self.path_raw = path_raw
        r"""str: path to the raw ``.hdf5`` files"""

        self.N_MCMC = N_MCMC
        r"""int: number of optimization procedures to run per posterior distribution"""
        self.T_MC = T_MC
        r"""int: total size of each optimization procedure"""
        self.T_BI = T_BI
        r"""int: duration of the `Burn-in` phase"""
        self.freq_save = freq_save
        r"""int: frequency of saved iterates, 1 means that all iterates were saved (used to show correct optimization procedure sizes in chain plots)"""

        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for results extraction"""

    @classmethod
    def read_estimator(
        cls,
        path_data_csv_out_optim_map: str,
        model_name: str,
    ) -> Tuple[np.ndarray, pd.DataFrame]:
        r"""reads the value of an already estimated MAP from a csv file.

        Parameters
        ----------
        path_data_csv_out_optim_map : str
            path to the csv file containing an already estimated MAP
        model_name : str
            name of the model, used to identify the posterior distribution

        Returns
        -------
        np.ndarray
            MAP estimator
        pd.DataFrame
            original DataFrame read from the csv file
        """
        path_file = f"{path_data_csv_out_optim_map}/"
        path_file += f"estimation_Theta_{model_name}_MAP_optim_map.csv"
        assert os.path.isfile(path_file), f"no MAP at {path_file}"

        df_results_map = pd.read_csv(path_file)
        df_results_map = df_results_map.sort_values(["n", "d"])

        N = df_results_map["n"].max() + 1
        D = df_results_map["d"].max() + 1

        Theta_MAP = np.zeros((N, D))
        for d in range(D):
            Theta_MAP[:, d] = df_results_map.loc[
                df_results_map["d"] == d, "Theta_MAP_optim_map"
            ].values

        return Theta_MAP, df_results_map

    def main(
        self,
        posterior: Posterior,
        model_name: str,
        scaler: Scaler,
        #
        list_idx_sampling: List[int],
        list_fixed_values: Union[List[float], np.ndarray],
        #
        estimator_plot: PlotsEstimator,
        Theta_true_scaled: Optional[np.ndarray] = None,
    ):
        r"""performs the data extraction, in this order:

        * step 1 : clppd
        * step 2 : kernel analysis
        * step 3 : objective evolution
        * step 4 : MAP estimator from samples
        * step 5 : model checking with Bayesian p-value

        Parameters
        ----------
        posterior : Posterior
            probability distribution. The goal of the optimization procedure was to find its mode, i.e., the minimum of its negative log pdf.
        model_name : str
            name of the model, used to identify the posterior distribution
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens) and its inverse
        list_idx_sampling : List[int]
            indices of the physical parameters that were sampled (the other ones were fixed)
        list_fixed_values : List[float]
            list of used values for the parameters fixed during the sampling
        estimator_plot : PlotsEstimator
            object used to plot the estimator figures
        Theta_true_scaled : Optional[np.ndarray], optional
            true value for the inferred physical parameter :math:`\Theta` (only possible for toy cases), by default None
        """
        list_mcmc_folders = [
            f"{x[0]}/mc_chains.hdf5"
            for x in os.walk(f"{self.path_raw}/{model_name}")
            if "optim_MAP_" in x[0]
        ]
        list_mcmc_folders.sort()

        N = posterior.N * 1
        D_sampling = posterior.D * 1
        D = len(list_fixed_values)
        L = posterior.L

        list_fixed_values_scaled = list(
            posterior.likelihood.forward_map.arr_fixed_values
        )
        list_fixed_values_scaled = [v if v != 0 else None for v in list_fixed_values]

        chain_type = "optim_map"

        # clppd
        ResultsCLPPD(
            model_name=model_name,
            chain_type=chain_type,
            path_img=self.path_img,
            path_data_csv_out=self.path_data_csv_out_optim_map,
            N_MCMC=self.N_MCMC,
            N=N,
            L=L,
        ).main(list_mcmc_folders, estimator_plot.map_shaper)

        # kernel analysis
        ResultsKernels(
            model_name,
            chain_type,
            self.path_img * 1,
            self.N_MCMC * 1,
            self.T_MC * 1,
            self.freq_save * 1,
        ).main(list_mcmc_folders)

        # objective evolution
        if Theta_true_scaled is not None:
            forward_map_evals = posterior.likelihood.evaluate_all_forward_map(
                Theta_true_scaled,
                compute_derivatives=False,
                compute_derivatives_2nd_order=False,
            )
            nll_utils = posterior.likelihood.evaluate_all_nll_utils(
                forward_map_evals,
                compute_derivatives=False,
                compute_derivatives_2nd_order=False,
            )
            objective_true = posterior.neglog_pdf(
                Theta_true_scaled, forward_map_evals, nll_utils
            )
        else:
            objective_true = None

        idx_lowest_obj, lowest_obj = ResultsObjective(
            model_name,
            chain_type,
            self.path_img * 1,
            self.N_MCMC * 1,
            self.T_MC * 1,
            self.T_BI * 1,
            self.freq_save * 1,
            N=N,
            D=D,
            L=L,
        ).main(list_mcmc_folders, objective_true)

        # MAP estimator from samples
        if Theta_true_scaled is not None:
            Theta_true_scaled_full = np.zeros((N, D))
            Theta_true_scaled_full[:, list_idx_sampling] += Theta_true_scaled

            for d in range(D):
                if list_fixed_values_scaled[d] is not None:
                    Theta_true_scaled_full[:, d] += list_fixed_values_scaled[d]
        else:
            Theta_true_scaled_full = None

        ResultsLowestObjective(
            model_name,
            chain_type,
            self.path_img,
            self.path_data_csv_out_optim_map,
            self.N_MCMC,
            self.T_MC,
            self.freq_save,
        ).main(
            list_mcmc_folders,
            lowest_obj,
            idx_lowest_obj,
            scaler,
            Theta_true_scaled_full,  # (N, D_sampling)
            list_idx_sampling,  # (D_sampling,)
            list_fixed_values,  # (D,)
            estimator_plot,
        )

        ResultsBayesPvalues(
            model_name=model_name,
            chain_type=chain_type,
            path_img=self.path_img,
            path_data_csv_out=self.path_data_csv_out_optim_map,
            N_MCMC=self.N_MCMC,
            N=N,
            D_sampling=D_sampling,
            plot_ESS=True,
        ).main(
            list_idx_sampling=list_idx_sampling,
            map_shaper=estimator_plot.map_shaper if N > 1 else None,
        )



import os
from typing import List, Optional, Tuple

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors

from beetroots.inversion.plots.map_shaper import MapShaper
from beetroots.inversion.results.utils.abstract_util import ResultsUtil


class ResultsCLPPD(ResultsUtil):

    __slots = (
        "model_name",
        "path_img",
        "path_data_csv_out",
        "N",
        "D",
    )

    def __init__(
        self,
        model_name: str,
        chain_type: str,
        path_img: str,
        path_data_csv_out: str,
        N_MCMC: int,
        N: int,
        L: int,
    ):
        assert chain_type in ["mcmc", "optim_map", "optim_mle"]

        self.model_name = model_name
        self.path_img = path_img
        self.path_data_csv_out = path_data_csv_out
        self.chain_type = chain_type

        self.N_MCMC = N_MCMC

        self.N = N
        self.L = L

    def read_data(
        self,
        list_chains_folders: List[str],
    ) -> Tuple[np.ndarray, np.ndarray]:
        clppd = np.zeros((self.N_MCMC, self.N, self.L))
        # p_values = np.zeros((self.N_MCMC, self.N, self.L))
        p_values_llh = np.zeros((self.N_MCMC, self.N))

        index = pd.MultiIndex.from_product(
            [list(range(self.N_MCMC)), list(range(self.N))],
            names=["seed", "n"],
        )
        df_clppd = pd.DataFrame(
            np.zeros((self.N_MCMC * self.N, self.L)),
            columns=np.arange(self.L),
            index=index,
        )
        # df_p_values = pd.DataFrame(
        #     np.zeros((self.N_MCMC * self.N, self.L)),
        #     columns=np.arange(self.L),
        #     index=index,
        # )
        df_p_values_llh = pd.DataFrame(
            np.zeros((self.N_MCMC * self.N,)),
            columns=["p-value-llh"],
            index=index,
        )

        for seed, mc_path in enumerate(list_chains_folders):
            with h5py.File(mc_path, "r") as f:
                clppd[seed] = np.array(f["clppd"])
                df_clppd.loc[(seed, slice(None)), :] = clppd[seed] * 1

                # p_values[seed] = np.array(f["p-values-y"])
                # df_p_values.loc[(seed, slice(None)), :] = p_values[seed] * 1

                p_values_llh[seed] = np.array(f["p-values-llh"])
                df_p_values_llh.loc[(seed, slice(None)), :] = p_values_llh[seed] * 1

        path_file = f"{self.path_data_csv_out}/"
        path_file += f"clppd_inter_{self.model_name}.csv"
        df_clppd.to_csv(path_file)

        # path_file = f"{self.path_data_csv_out}/"
        # path_file += f"p_value_inter_{self.model_name}.csv"
        # df_p_values.to_csv(path_file)

        path_file = f"{self.path_data_csv_out}/"
        path_file += f"p_value_llh_{self.model_name}.csv"
        df_p_values_llh.to_csv(path_file)

        return clppd, p_values_llh  # p_values,

    def create_folders(self) -> str:
        folder_path_inter = f"{self.path_img}/clppd"
        folder_path = f"{folder_path_inter}/{self.model_name}"
        for path_ in [folder_path_inter, folder_path]:
            if not os.path.isdir(path_):
                os.mkdir(path_)

        return folder_path

    def main(
        self,
        list_chains_folders: List[str],
        map_shaper: Optional[MapShaper],
    ) -> None:

        if self.N < 1:
            msg = "this function should only be called when N > 1 "
            msg += "to avoid 1-pixel maps"
            raise ValueError(msg)

        clppd, p_values_llh = self.read_data(list_chains_folders)  # p_values,
        folder_path = self.create_folders()

        print("starting clppd plots")

        clppd = -2 / self.L * np.sum(np.log(clppd), axis=2)  # (N_MCMC, N)

        if map_shaper is not None:
            for seed in range(self.N_MCMC):
                clppd_seed_shaped = map_shaper.from_vector_to_map(clppd[seed])

                plt.figure(figsize=(8, 6))
                if self.N_MCMC == 1:
                    plt.title("CLPPD")
                else:
                    plt.title(f"CLPPD for seed={seed}")

                plt.imshow(
                    clppd_seed_shaped,
                    norm=colors.LogNorm(),
                    cmap="jet",
                    origin="lower",
                )
                plt.colorbar()
                # plt.tight_layout()
                plt.savefig(
                    f"{folder_path}/clppd_{self.chain_type}_seed{seed}.PNG",
                    bbox_inches="tight",
                )
                plt.close()

            # # * p-values computed with y^{rep}_\ell <= y_\ell
            # p_values = np.min(p_values, axis=2)  # (N_MCMC, N)
            # assert p_values.shape == (
            #     self.N_MCMC,
            #     self.N,
            # ), f"should be {(self.N_MCMC, self.N)}, is {p_values.shape}"

            # if map_shaper is not None:
            #     for seed in range(self.N_MCMC):
            #         p_values_seed_shaped = map_shaper.from_vector_to_map(
            #             p_values[seed],
            #         )

            #         plt.figure(figsize=(8, 6))
            #         title = "p-values"
            #         if self.N_MCMC > 1:
            #             title += f" for seed={seed}"
            #         plt.title(title)

            #         plt.imshow(
            #             p_values_seed_shaped,
            #             norm=colors.LogNorm(),
            #             cmap="jet",
            #             origin="lower",
            #         )
            #         plt.colorbar()
            #         # plt.tight_layout()
            #         filename = f"{folder_path}/p_values_{self.chain_type}"
            #         filename += f"_seed{seed}.PNG"
            #         plt.savefig(filename, bbox_inches="tight")
            #         plt.close()

            print("clppd plots done")
        return



import abc
import datetime
import json
import os
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from beetroots.approx_optim.approach_type import utils


class ApproxParamsOptim(abc.ABC):
    r"""Optimization of an approximation parameter for the likelihood function ``beetroots.modelling.likelihoods.approx_censored_add_mult.MixingModelsLikelihood``.
    The optimization framework used to adjust this parameter is introduced in Appendix A of :cite:t:`paludEfficientSamplingNon2023`.
    """

    def __init__(
        self,
        list_lines: List[str],
        name: str,
        D: int,
        D_no_kappa: int,
        K: int,
        log10_f_grid_size: int,
        N_samples_y: int,
        max_workers: int,
        sigma_a: Union[np.ndarray, float],
        sigma_m: Union[np.ndarray, float],
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
    ):
        r"""

        Parameters
        ----------
        list_lines : List[str]
            names of the observables for which the likelihood parameter needs to be adjusted
        name : str
            name of the process, to be used in the output folder name
        D : int
            total number of physical parameters involved in the forward map
        D_no_kappa : int
            total number of physical parameters involved in the forward map except for the scaling parameter :math:`\kappa` (if it is part of the considered physical parameters)
        K : int
            the number of sampled theta values is :math:`K^D`
        log10_f_grid_size : int
            number of points in the grid on :math:`\log f_\ell(\theta)`
        N_samples_y : int
            number of samples for :math:`y_\ell`
        max_workers : int
            maximum number of workers that can be used for optimization or results extraction
        sigma_a : Union[np.ndarray, float]
            standard deviation of the additive Gaussian noise
        sigma_m : Union[np.ndarray, float]
            standard deviation parameter of the multiplicative lognormal noise
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        """
        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for optimization or results extraction"""

        self.list_lines = list_lines
        r"""List[str]: names of the observables for which the likelihood parameter needs to be adjusted"""
        self.L = len(list_lines)

        if isinstance(sigma_a, np.ndarray):
            assert len(sigma_a.shape) == 2, f"{sigma_a.shape}"
            assert sigma_a.shape[1] == self.L
            N = sigma_a.shape[0]
        else:
            N = 1
            sigma_a = sigma_a * np.ones((N, self.L))

        if isinstance(sigma_m, np.ndarray):
            assert sigma_m.shape == (N, self.L), f"{sigma_m.shape}"
        else:
            N = 1
            sigma_m = sigma_m * np.ones((N, self.L))

        self.N = N
        r"""int: number of pixels / components for which the optimization needs to be performed"""

        assert isinstance(sigma_a, np.ndarray)
        assert isinstance(sigma_m, np.ndarray)
        self.sigma_a = sigma_a
        r"""np.ndarray: standard deviations of the additive Gaussian noise"""
        self.sigma_m = sigma_m
        r"""np.ndarray: standard deviation parameter of the multiplicative lognormal noise"""

        self.D = D
        r"""int: total number of physical parameters involved in the forward map"""

        self.D_no_kappa = D_no_kappa
        r"""int: total number of physical parameters involved in the forward map except for the scaling parameter :math:`\kappa` (if it is part of the considered physical parameters)"""

        self.K = K
        r"""int: the number of sampled theta values is ``K^D_sampling``"""
        self.N_samples_y = N_samples_y
        r"""int: number of samples for :math:`y_\ell`"""
        self.log10_f_grid_size = log10_f_grid_size  # nb points P(theta) grid
        r"""int: number of points in the grid on :math:`\log_{10} f_\ell(\theta)`"""

        self.setup_plot_text_sizes(small_size, medium_size, bigger_size)
        self.create_empty_output_folders(name)

    def setup_plot_text_sizes(
        self,
        small_size: int = 16,
        medium_size: int = 20,
        bigger_size: int = 24,
    ) -> None:
        r"""defines text sizes on matplotlib plots

        Parameters
        ----------
        small_size : int, optional
            size for basic text, axes titles, xticks and yticks, by default 16
        medium_size : int, optional
            size of the axis labels, by default 20
        bigger_size : int, optional
            size of the figure title, by default 24
        """
        plt.rc("font", size=small_size)  # controls default text sizes
        plt.rc("axes", titlesize=small_size)  # fontsize of the axes title
        plt.rc("axes", labelsize=medium_size)  # fontsize of the x and y labels
        plt.rc("xtick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("ytick", labelsize=small_size)  # fontsize of the tick labels
        plt.rc("legend", fontsize=small_size)  # legend fontsize
        plt.rc("figure", titlesize=bigger_size)  # fontsize of the figure title
        return

    def create_empty_output_folders(self, name: str) -> None:
        r"""creates the directories that receive the results of the likelihood parameter optimization

        Parameters
        ----------
        name : str
            name of the simulation to be run
        """
        now = datetime.datetime.now()
        dt_str = now.strftime("%Y-%m-%d_%H")

        # path to the outputs dir
        path_ouput_general = os.path.abspath(
            f"{os.path.abspath(__file__)}/../../../outputs"
        )

        self.path_output_sim = os.path.abspath(
            f"{path_ouput_general}/approx_optim_{name}_{dt_str}"
        )
        r"""str: path to the output root folder, e.g., ``./outputs/simu1``"""

        self.path_img = self.path_output_sim + "/img"
        r"""str: path to the output image folder, e.g., ``./outputs/simu1/img``"""
        self.path_img_hist = self.path_img + "/hist"
        r"""str: path to the histogram image folder, e.g., ``./outputs/simu1/img/hist``"""
        self.path_img_hist_final = self.path_img + "/hist_final"
        r"""str: path to the final histogram image folder, e.g., ``./outputs/simu1/img/hist_final``"""
        self.path_img_final = self.path_img + "/final"
        r"""str: path to the final image folder, e.g., ``./outputs/simu1/img/final``"""
        self.path_logs = self.path_output_sim + "/logs"
        r"""str: path to the procedure logs, e.g., ``./outputs/logs``"""
        self.path_params = self.path_output_sim + "/optim_params"
        r"""str: path to the adjusted likelihood parameters, e.g., ``./outputs/optim_params``"""

        for folder_path in [
            path_ouput_general,
            self.path_output_sim,
            self.path_img,
            self.path_img_hist,
            self.path_img_hist_final,
            self.path_img_final,
            self.path_logs,
            self.path_params,
        ]:
            if not os.path.isdir(folder_path):
                os.mkdir(folder_path)

        return

    def sample_theta(
        self,
        lower_bounds: np.ndarray,
        upper_bounds: np.ndarray,
    ) -> np.ndarray:
        r"""sample :math:`\theta` from Stratified MC in cube

        Parameters
        ----------
        K : int
            total number of samples per axis
        lower_bounds : np.ndarray of shape (D,)
            lower bounds of cube on :math:`\theta`
        upper_bounds : np.ndarray of shape (D,)
            upper bounds of cube on :math:`\theta`

        Returns
        -------
        np.ndarray of shape (N_samples, D)
            :math:`\theta` samples
        """
        print("starting generation of theta values with Stratified MC")
        rng = np.random.default_rng(42)
        D_sampling: int = upper_bounds.size

        delta_upper_lower = np.array(
            [upper_bounds[d] - lower_bounds[d] for d in range(D_sampling)]
        )
        list_lower_bounds = [
            np.arange(self.K) / self.K * delta_upper_lower[d] + lower_bounds[d]
            for d in range(D_sampling)
        ]
        list_lower_bounds = np.meshgrid(*list_lower_bounds)
        list_lower_bounds = [
            lower_bounds_Theta_i.flatten() for lower_bounds_Theta_i in list_lower_bounds
        ]
        list_samples = []
        for d, lower_bounds_Theta_i in enumerate(list_lower_bounds):
            Vd = rng.uniform(
                low=lower_bounds_Theta_i,
                high=lower_bounds_Theta_i + delta_upper_lower[d] / self.K,
            )
            list_samples.append(Vd)
        x = np.vstack(list_samples).T
        print("generation of theta values with Stratified MC done")
        return x

    def plot_hist_log10_f_Theta(
        self,
        log10_f_Theta: np.ndarray,
        log10_f_Theta_low: float,
        log10_f_Theta_high: float,
        list_log10_f_grid: np.ndarray,
        pdf_kde_log10_f_Theta: np.ndarray,
        ell: int,
    ) -> None:
        r"""plots histogram of :math:`log_{10}(f_\ell(\theta))`

        Parameters
        ----------
        log10_f_Theta : np.ndarray of shape (-1, 1)
            array of values of :math:`\log_{10} f_\ell (\theta)` for considered line
        log10_f_Theta_low: float
            lower bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        log10_f_Theta_high: float
            upper bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        list_log10_f_grid: np.ndarray
            grid values of :math:`\log_{10} f_\ell (\theta)` for the considered line
        pdf_kde_log10_f_Theta: np.ndarray
            pdf of :math:`\log_{10} f_\ell (\theta)` evaluated with a kernel density estimator
        ell : int
            index of the line
        """
        plt.figure(figsize=(8, 6))
        plt.title(f"line {ell} : {self.list_lines[ell]}")
        plt.hist(
            log10_f_Theta,
            bins=self.log10_f_grid_size,
            range=(log10_f_Theta_low, log10_f_Theta_high),
            density=True,
            label="samples",
        )
        plt.plot(list_log10_f_grid, pdf_kde_log10_f_Theta, "k--", label="KDE")
        plt.xlabel(r"$\log f_\ell (\theta)$")
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.savefig(f"{self.path_img_hist}/hist_log_f_Theta_{self.list_lines[ell]}.PNG")
        plt.close()

        return

    def plot_hist_log10_f_Theta_with_optim_results(
        self,
        log10_f_Theta: np.ndarray,
        log10_f_Theta_low: float,
        log10_f_Theta_high: float,
        list_log10_f_grid: np.ndarray,
        pdf_kde_log10_f_Theta: np.ndarray,
        n: int,
        ell: int,
        best_point: np.ndarray,
    ) -> None:
        r"""plots histogram of :math:`\log_{10} f_\ell (\theta)`

        Parameters
        ----------
        log10_f_Theta : np.ndarray of shape (-1, 1)
            array of values of :math:`\log_{10} f_\ell (\theta)` for considered line
        log10_f_Theta_low: float
            lower bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        log10_f_Theta_high: float
            upper bound for :math:`\log_{10} f_\ell (\theta)` for the considered line
        list_log10_f_grid: np.ndarray
            grid values of :math:`\log_{10} f_\ell (\theta)` for the considered line
        pdf_kde_log10_f_Theta: np.ndarray
            pdf of :math:`\log_{10} f_\ell (\theta)` evaluated with a kernel density estimator
        n : int
            pixel / component index
        ell : int
            index of the line
        best_point : np.ndarray
            position for the best point, to be displayed
        """
        lambda_ = 1 - utils.compute_lambda(
            a0=best_point[0],
            a1=best_point[1],
            f_Theta_true=np.exp(list_log10_f_grid * np.log(10.0)),
        )

        var_eps_m = np.exp(self.sigma_m**2) * (np.exp(self.sigma_m**2) - 1)
        log10_f0_n_ell = (
            0.5
            * (2 * np.log(self.sigma_a[n, ell]) - np.log(var_eps_m[n, ell]))
            / np.log(10)
        )

        plt.figure(figsize=(8, 6))
        plt.title(f"line {ell} : {self.list_lines[ell]}")
        vals, _, _ = plt.hist(
            log10_f_Theta,
            bins=self.log10_f_grid_size,
            range=(log10_f_Theta_low, log10_f_Theta_high),
            density=True,
            label="samples",
        )
        plt.plot(list_log10_f_grid, pdf_kde_log10_f_Theta, "k--", label="KDE")
        plt.plot(
            list_log10_f_grid,
            lambda_ * vals.max(),
            "r-",
            label=r"multi. weight $\lambda$",
        )
        plt.axvline(
            np.log10(self.sigma_a[n, ell]), c="orange", ls="-", label=r"$\sigma_a$"
        )
        plt.axvline(
            log10_f0_n_ell,
            c="orange",
            ls="--",
            label=r"$\sigma_a^2 = f_\ell(\theta)^2$ Var$( \epsilon^{(m)}_{\ell} )$",
        )

        plt.xlabel(r"$\log_{10} f_\ell (\theta)$")
        plt.grid()
        plt.legend()
        plt.tight_layout()
        plt.savefig(
            f"{self.path_img_hist_final}/hist_final_log_f_Theta_n{n}_{self.list_lines[ell]}.PNG"
        )
        plt.close()

        return

    def save_setup_to_json(
        self, n: int, ell: int, pbounds: Dict[str, np.ndarray]
    ) -> None:
        """save optimization context and parameters

        Parameters
        ----------
        n : int
            pixel / component index
        ell : int
            observable index
        pbounds : dict[str, np.ndarray]
            contains the bounds on the parameters to be adjusted
        """
        optim_params = {
            "n": n,
            "ell": ell,
            "line": self.list_lines[ell],
            "sigma_a": self.sigma_a[n, ell],
            "sigma_m": self.sigma_m[n, ell],
            "K": self.K,
            "N_samples_theta": self.N_samples_theta,
            "N_samples_y": self.N_samples_y,
            "log10_f_grid_size": self.log10_f_grid_size,
            "a0": {"low": pbounds["a0"][0], "high": pbounds["a0"][1]},
            "a1": {"low": pbounds["a1"][0], "high": pbounds["a1"][1]},
        }
        with open(
            f"{self.path_params}/optim_params_n{n}_{self.list_lines[ell]}.json",
            "w",
            encoding="utf-8",
        ) as f:
            json.dump(
                optim_params,
                f,
                ensure_ascii=False,
                indent=4,
            )
        return

    def rewrite_logs_correct_json_format(self) -> None:
        r"""rewrites the log files with correct json format"""
        for ell in range(self.L):
            line = self.list_lines[ell] * 1

            for n in range(self.N):
                with open(f"{self.path_logs}/logs_n{n}_{line}.json", "r") as f:
                    text = f.read()

                text = "[" + text.replace("}}", "}},")[:-2] + "]"
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{line}.json",
                    "w",
                ) as f:
                    f.write(text)

        return

    def extract_optimal_params(self) -> pd.DataFrame:
        """extracts the adjusted likelihood parameters from the log files and gather them in a DataFrame

        Returns
        -------
        pd.DataFrame
            DataFrame with the set of evaluated optimal parameters for each component n and line ell
        """
        self.rewrite_logs_correct_json_format()

        list_best = []
        for n in range(self.N):
            for ell in range(self.L):
                line = self.list_lines[ell] * 1
                # import data
                with open(
                    f"{self.path_logs}/logs_correct_format_n{n}_{line}.json",
                    "r",
                ) as f:
                    data = json.load(f)

                # Flatten data
                df = pd.json_normalize(data, max_level=1)
                df = df.rename(columns={"params.a0": "a0", "params.a1": "a1"})

                # get best
                df_sorted = df.sort_values("target", ascending=False)
                idx_best = df_sorted.index[0]
                dict_best = {
                    "n": n,
                    "ell": ell,
                    "line": self.list_lines[ell],
                    "a0_best": df.at[idx_best, "a0"],
                    "a1_best": df.at[idx_best, "a1"],
                    "target_best": df.at[idx_best, "target"],
                }
                list_best.append(dict_best)

        # * save best points
        df_best = pd.DataFrame(list_best)
        df_best.to_csv(f"{self.path_output_sim}/best_params.csv", index=False)
        return df_best

    def setup_params_bounds(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, float, float]:
        r"""sets the bounds on the parameters to be adjusted, defined here as transition interval center for ``a0`` position and width for ``a1``

        Returns
        -------
        log10_f0 : np.ndarray of shape (N, L)
            values of :math:`f_\ell(\theta)` at which additive and multiplicative noise variances are equal
        bounds_a0_low : np.ndarray of shape (N, L)
            lower bounds on the center of the transition interval (defined as deltas around the ``log10_f0``)
        bounds_a0_high : np.ndarray of shape (N, L)
            upper bounds on the center of the transition interval (defined as deltas around the ``log10_f0``)
        bounds_a1_low : float
            lower value for the transition interval size
        bounds_a1_high : float
            upper value for the transition interval size
        """
        print("starting definition of parameters bounds")
        # log10_f0 = log10(ratio std of noises)
        # ie value of f at which noise variances are equal
        var_eps_m: np.ndarray = np.exp(self.sigma_m**2) * (
            np.exp(self.sigma_m**2) - 1
        )
        log10_f0: np.ndarray = (
            0.5 * (2 * np.log(self.sigma_a) - np.log(var_eps_m)) / np.log(10)
        )

        bounds_a0_low: np.ndarray = log10_f0 - 2  # (N, L)
        bounds_a0_high: np.ndarray = log10_f0 + 8  # (N, L)

        bounds_a1_low = 0.01
        bounds_a1_high = 2.0
        print("definition of parameters bounds done")

        return (
            log10_f0,
            bounds_a0_low,
            bounds_a0_high,
            bounds_a1_low,
            bounds_a1_high,
        )



r"""Contains a class of sampler used in the Meudon PDR code Bayesian inversion problems
"""
import copy
from typing import Optional, Tuple, Union

import numpy as np
from scipy.special import softmax
from tqdm.auto import tqdm

from beetroots.modelling.likelihoods.abstract_likelihood import Likelihood
from beetroots.modelling.posterior import Posterior
from beetroots.sampler.abstract_sampler import Sampler
from beetroots.sampler.saver.abstract_saver import Saver
from beetroots.sampler.utils import utils
from beetroots.sampler.utils.mml import EBayesMMLELogRate
from beetroots.sampler.utils.my_sampler_params import MySamplerParams


class MySampler(Sampler):
    r"""Defines the sampler proposed in :cite:t:`paludEfficientSamplingNon2023` that randomly combines two transition kernels :

    1. a independent MTM-chromatic Gibbs transition kernel
    2. a position-dependent MALA transition kernel with the RMSProp pre-conditioner
    """

    def __init__(
        self,
        my_sampler_params: MySamplerParams,
        D: int,
        L: int,
        N: int,
        rng: np.random.Generator = np.random.default_rng(42),
    ):
        r"""

        Parameters
        ----------
        my_sampler_params : MySamplerParams
            contains the main parameters of the algorithm
        D : int
            total number of physical parameters to reconstruct
        L : int
            number of observables per component :math:`n`
        N : int
            total number of pixels to reconstruct
        rng : numpy.random.Generator, optional
            random number generator (for reproducibility), by default np.random.default_rng(42)
        """

        # P-MALA params
        # ! redefine size of params
        self.eps0 = my_sampler_params.initial_step_size
        r"""float: step size used in the Position-dependent MALA transition kernel, denoted :math:`\epsilon > 0` in the article"""

        self.lambda_ = my_sampler_params.extreme_grad
        r"""float: limit value that avoids division by zero when computing the RMSProp preconditioner, denoted :math:`\eta > 0` in the article"""

        self.alpha = my_sampler_params.history_weight
        r"""float: weight of past values of :math:`v` in the exponential decay (cf RMSProp preconditioner), denoted :math:`\alpha \in ]0,1[` in the article"""

        # MTM params
        # assert np.isclose(
        #     int(pow(my_sampler_params.k_mtm, 1 / D)) ** D, my_sampler_params.k_mtm
        # ), "number of candidates for mtm needs to have an integer D-root"
        self.k_mtm = my_sampler_params.k_mtm
        r"""int: number of candidates in the MTM kernel, denoted :math:`K` in the article"""

        # overall
        self.selection_probas = my_sampler_params.selection_probas
        r"""np.ndarray: vector of selection probabilities for the MTM and PMALA kernels, respectively, i.e., :math:`[p_{MTM}, 1 - p_{MTM}]`"""
        assert (
            np.sum(self.selection_probas) == 1
        ), f"{self.selection_probas} should sum to 1"

        self.stochastic = my_sampler_params.is_stochastic
        r"""bool: if True, the algorithm performs sampling, and optimization otherwise"""

        self.compute_correction_term = my_sampler_params.compute_correction_term
        r"""bool: wether or not to use the correction term (denoted :math:`\gamma` in the article) during the sampling (only used if `is_stochastic=True`)"""

        self.compute_derivatives_2nd_order = (
            self.stochastic and self.compute_correction_term
        )
        r"""bool: wether to compute the expensive second order derivatives terms. Only true when the sampler runs a Markov chain (2nd order never used in optimization) and when the correction term denoted :math:`\gamma` is to be computed."""

        self.D = D
        r"""int: total number of physical parameters to reconstruct"""
        self.L = L
        r"""int: number of observables per component :math:`n`"""
        self.N = N
        r"""int: total number of pixels to reconstruct"""

        self.rng = rng
        r"""numpy.random.Generator: random number generator (for reproducibility)"""

        # initialization values, not to be kept during sampling
        self.v = np.zeros((N * D,))
        r"""np.ndarray: RMSProp gradient variance vector, denoted :math:`v` in the article"""

        self.current = {}
        r"""dict: contains all the data about the current iterate (including the evaluations of the forward map and derivatives, etc.)"""

    def generate_random_start_Theta_1pix(
        self, Theta: np.ndarray, posterior: Posterior, idx_pix: np.ndarray
    ) -> np.ndarray:
        r"""draws a random vectors for components :math:`n` (e.g., a pixel :math:`\theta_n`). The distribution used to draw these vectors is:

        * the smooth indicator prior
        * a combination of the smooth indicator prior and of a Gaussian mixture defined with the set of all combinations of neighbors of component :math:`n`

        Parameters
        ----------
        Theta : np.ndarray
            current iterate
        posterior : Posterior
            contains the lower and upper bounds of the hypercube
        idx_pix : np.ndarray
            indices of the pixels

        Returns
        -------
        np.array of shape (n_pix, self.k_mtm, D)
            random element of the hypercube defined by the lower and upper bounds with uniform distribution

        Raises
        ------
        ValueError : if ``posterior.prior_indicator`` is None
        """
        seed = self.rng.integers(0, 1_000_000_000)
        n_pix = idx_pix.size

        if posterior.prior_indicator is None:
            raise ValueError("The Posterior has no specified smooth indicator prior")

        if posterior.prior_spatial is None:
            # * sample from smooth indicator prior
            return utils.sample_smooth_indicator(
                posterior.prior_indicator.lower_bounds,
                posterior.prior_indicator.upper_bounds,
                posterior.prior_indicator.indicator_margin_scale,
                size=(n_pix * self.k_mtm, self.D),
                seed=seed,
            ).reshape((n_pix, self.k_mtm, self.D))

        else:
            return utils.sample_conditional_spatial_and_indicator_prior(
                Theta,
                posterior.prior_spatial.list_edges,
                posterior.prior_spatial.weights,
                posterior.prior_indicator.lower_bounds,
                posterior.prior_indicator.upper_bounds,
                posterior.prior_indicator.indicator_margin_scale,
                idx_pix=idx_pix,
                k_mtm=self.k_mtm,
                seed=seed,
            )  # (n_pix, self.k_mtm, D)

    def _update_model_check_values(
        self,
        dict_model_check: dict,
        likelihood: Likelihood,
        nll_full: np.ndarray,
        objective: float,
    ) -> dict:
        count_pval = dict_model_check["count_pval"] * 1
        y_copy = likelihood.y * 1

        if self.stochastic:
            dict_model_check["clppd_online"] *= count_pval / (count_pval + 1)
            dict_model_check["clppd_online"] += np.exp(-nll_full) / (count_pval + 1)

            y_rep = likelihood.sample_observation_model(
                self.current["forward_map_evals"],
                self.rng,
            )
            likelihood_rep = copy.deepcopy(likelihood)
            likelihood_rep.y = y_rep * 1

            assert np.allclose(likelihood.y, y_copy), "nooooo"

            nll_utils_rep = likelihood_rep.evaluate_all_nll_utils(
                self.current["forward_map_evals"],
                idx=None,
                compute_derivatives=False,
            )
            nll_y_rep_full = likelihood_rep.neglog_pdf(
                self.current["forward_map_evals"],
                nll_utils_rep,
                full=True,
            )

            # p-value per (N, L) with y_rep_{n,ell} <= y_{n,ell}
            dict_model_check["p_values_y"] *= count_pval / (count_pval + 1)
            dict_model_check["p_values_y"] += (y_rep <= likelihood.y) / (count_pval + 1)

            # p-value per (N,) with
            # p(y_rep_n \vert theta_n) <= p(y_n \vert theta_n)
            nll_y = np.sum(nll_full, axis=1)  # (N,)
            nll_y_rep = np.sum(nll_y_rep_full, axis=1)  # (N,)

            dict_model_check["p_values_llh"] *= count_pval / (count_pval + 1)
            dict_model_check["p_values_llh"] += (nll_y_rep >= nll_y) / (count_pval + 1)

            dict_model_check["count_pval"] += 1

        else:
            if objective < dict_model_check["best_objective"]:
                dict_model_check["best_objective"] = objective * 1
                dict_model_check["clppd_online"] = np.exp(-nll_full)

            # p-values are computed at the end of the optimisation process.

        return dict_model_check

    def _finalize_model_check_values(
        self,
        dict_model_check: dict,
        likelihood: Likelihood,
        forward_map_evals: dict,
        nll_full: np.ndarray,
    ) -> dict:
        if not self.stochastic:
            # optimization p-value computations on estimated \hat{\theta}
            for count_pval in tqdm(range(self.ESS_OPTIM)):
                y_rep = likelihood.sample_observation_model(
                    forward_map_evals,
                    self.rng,
                )
                likelihood_rep = copy.deepcopy(likelihood)
                likelihood_rep.y = y_rep * 1
                nll_utils_rep = likelihood_rep.evaluate_all_nll_utils(
                    forward_map_evals,
                    idx=None,
                    compute_derivatives=False,
                )
                nll_y_rep_full = likelihood_rep.neglog_pdf(
                    forward_map_evals,
                    nll_utils_rep,
                    full=True,
                )

                # p-value per (N, L) with y_rep_{n,ell} <= y_{n,ell}
                dict_model_check["p_values_y"] *= count_pval / (count_pval + 1)
                dict_model_check["p_values_y"] += (y_rep <= likelihood.y) / (
                    count_pval + 1
                )

                # p-value per (N,) with
                # p(y_rep_n \vert theta_n) <= p(y_n \vert theta_n)
                nll_y = np.sum(nll_full, axis=1)  # (N,)
                nll_y_rep = np.sum(nll_y_rep_full, axis=1)  # (N,)

                dict_model_check["p_values_llh"] *= count_pval / (count_pval + 1)
                dict_model_check["p_values_llh"] += (nll_y_rep >= nll_y) / (
                    count_pval + 1
                )

        # this p-value should be between 0 and 0.5
        dict_model_check["p_values_y"] = np.where(
            dict_model_check["p_values_y"] > 0.5,
            1 - dict_model_check["p_values_y"],
            dict_model_check["p_values_y"],
        )

        return dict_model_check

    def sample(
        self,
        posterior: Posterior,
        saver: Saver,
        max_iter: int,
        Theta_0: Optional[np.ndarray] = None,
        disable_progress_bar: bool = False,
        #
        regu_spatial_N0: Union[int, float] = np.infty,
        regu_spatial_scale: float = 1.0,
        regu_spatial_vmin: float = 1e-8,
        regu_spatial_vmax: float = 1e8,
        #
        T_BI: int = 0,  # used only for clppd
    ) -> None:
        r"""main method of the class, runs the sampler

        Parameters
        ----------
        posterior : Posterior
            probability distribution to be sampled
        saver : Saver
            object responsible for progressively saving the Markov chain data during the run
        max_iter : int
            total duration of a Markov chain
        Theta_0 : Optional[np.ndarray], optional
            starting point, by default None
        disable_progress_bar : bool, optional
            wether to disable the progress bar, by default False
        regu_spatial_N0 : Union[int, float], optional
            number of iterations defining the initial update phase (for spatial regularization weight optimization). np.infty means that the optimization phase never starts, and that the weight optimization is not applied. by default np.infty
        regu_spatial_scale : Optional[float], optional
            scale parameter involved in the definition of the projected gradient
            step size (for spatial regularization weight optimization). by default 1.0
        regu_spatial_vmin : Optional[float], optional
            lower limit of the admissible interval (for spatial regularization weight optimization), by default 1e-8
        regu_spatial_vmax : Optional[float], optional
            upper limit of the admissible interval (for spatial regularization weight optimization), by default 1e8
        T_BI : int, optional
            duration of the `Burn-in` phase, by default 0
        """
        additional_sampling_log = {}

        if Theta_0 is None:
            print("starting from a random point")
            Theta_0 = self.generate_random_start_Theta(posterior)  # (N, D)

        assert Theta_0 is not None
        assert Theta_0.shape == (self.N, self.D)

        self.current = posterior.compute_all(
            Theta_0, compute_derivatives_2nd_order=self.compute_derivatives_2nd_order
        )

        assert np.isnan(self.current["objective"]) == 0
        assert np.sum(np.isnan(self.current["grad"])) == 0
        # assert (
        #     self.current["forward_map_evals"]["f_Theta"].min() >= 0
        # ), f"{self.current['forward_map_evals']['f_Theta'].min()}"  # {self.current['forward_map_evals']['log_f_Theta'].min()}"

        # if v0 is None:
        #     v_max = (self.current["grad"] ** 2).max(axis=0)
        #     self.v = (v_max[None, :] * np.ones((self.N, self.D))).flatten()
        #     assert np.sum(np.isnan(self.v)) == 0, np.sum(np.isnan(self.v))
        # else:
        #     self.v = v0
        self.v = self.current["grad"].flatten() ** 2
        assert np.sum(np.isnan(self.v)) == 0.0
        assert np.sum(np.isinf(self.v)) == 0.0

        # print(self.v, Theta_0, self.lambda_)

        # self.u = self.current["grad"].flatten() * self.current["hess_diag"].flatten()
        # assert self.u.shape == (self.N * self.D,)

        # if sample_regu_weights and T_BI_reguweights is None:
        #     T_BI_reguweights = 0
        # if not (sample_regu_weights) and T_BI_reguweights is None:
        #     T_BI_reguweights = max_iter * 1

        rng_state_array, _ = self.get_rng_state()

        # self.j_t = 0
        self.j_t = np.zeros((self.N * self.D,))

        # if self.N > 10:
        #     print(f"at start: obj = {self.current['objective']}")

        # n_sites = len(posterior.prior_spatial.dict_sites)

        # n_repetitions_first_mtm = 1
        # list_n_first_samples = list(range(n_sites)) * n_repetitions_first_mtm
        # self.rng.shuffle(list_n_first_samples)

        # if self.N > 1000:
        # list_n_first_samples = []
        # list_n_first_samples = list_n_first_samples[:250]

        regu_weights_optimizer = EBayesMMLELogRate(
            scale=regu_spatial_scale,
            N0=regu_spatial_N0,
            N1=+np.infty,
            dim=self.D * self.N,
            vmin=regu_spatial_vmin,
            vmax=regu_spatial_vmax,
            homogeneity=2.0,
            exponent=0.8,
        )
        optimize_regu_weights = regu_weights_optimizer.N0 < np.infty

        # clppd = computed log point-wise predictive density.
        # if self.stochastic : avg of all pred. likelihood terms (with burn-in)
        # but burn-in values are negligible (0) compared to non burn-in
        # else : predictive likelihood with best param theta
        clppd_online = np.zeros((self.N, self.L))
        # utilitary variables
        best_objective = np.infty  # used only if not self.stochastic
        count_pval = 0  # used only if self.stochastic

        # p(y^{rep}_\ell <= y_\ell | y)
        p_values_y = np.zeros((self.N, self.L))
        # p(y^{rep}_\ell \in [ q_{25\%}(y_\ell), q_{75\%}(y_\ell) ] | y)
        p_values_llh = np.zeros((self.N,))

        dict_model_check = {
            "clppd_online": clppd_online,
            "best_objective": best_objective,
            "count_pval": count_pval,
            "p_values_y": p_values_y,
            "p_values_llh": p_values_llh,
        }

        for t in tqdm(range(1, max_iter + 1), disable=disable_progress_bar):
            if optimize_regu_weights and (self.N > 1):
                assert posterior.prior_spatial is not None

                if t >= regu_weights_optimizer.N0:
                    tau_t = self.sample_regu_hyperparams(
                        posterior,
                        regu_weights_optimizer,
                        t,
                        self.current["Theta"] * 1,
                    )

                    posterior.prior_spatial.weights = tau_t * 1

                    # recompute posterior neg log pdf and gradients with
                    # new spatial regularization parameter
                    self.current = posterior.compute_all(
                        self.current["Theta"],
                        self.current["forward_map_evals"],
                        self.current["nll_utils"],
                        compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
                    )

                additional_sampling_log["tau"] = posterior.prior_spatial.weights * 1
            # ------

            type_t = np.argmax(
                self.rng.multinomial(
                    1,
                    pvals=self.selection_probas,
                )
            )
            if type_t == 0:
                (
                    accepted_t,
                    log_proba_accept_t,
                ) = self.generate_new_sample_mtm(t, posterior)
            else:
                assert type_t == 1
                (
                    accepted_t,
                    log_proba_accept_t,
                ) = self.generate_new_sample_pmala_rmsprop(t, posterior)

            # # check for nan in Theta
            # if np.sum(np.isnan(self.current["Theta"])) > 0:
            #     print(f"type_t : {type_t}")
            #     raise ValueError(
            #         f"Theta contains {np.sum(np.isnan(self.current['Theta']))} nan"
            #     )

            # * if the memory is empty : initialize it
            if saver.memory == {}:
                additional_sampling_log["v"] = self.v.reshape((self.N, self.D)) * 1
                additional_sampling_log["type_t"] = type_t
                additional_sampling_log["accepted_t"] = accepted_t
                additional_sampling_log["log_proba_accept_t"] = log_proba_accept_t

                dict_objective, nll_full = posterior.compute_all_for_saver(
                    self.current["Theta"],
                    self.current["forward_map_evals"],
                    self.current["nll_utils"],
                )

                if t > T_BI:
                    dict_model_check = self._update_model_check_values(
                        dict_model_check,
                        posterior.likelihood,
                        nll_full,
                        dict_objective["objective"] * 1,
                    )

                saver.initialize_memory(
                    max_iter,
                    t,
                    Theta=self.current["Theta"],
                    forward_map_evals=self.current["forward_map_evals"],
                    nll_utils=self.current["nll_utils"],
                    dict_objective=dict_objective,
                    additional_sampling_log=additional_sampling_log,
                )

                rng_state_array, rng_inc_array = self.get_rng_state()

                saver.update_memory(
                    t,
                    Theta=self.current["Theta"],
                    forward_map_evals=self.current["forward_map_evals"],
                    nll_utils=self.current["nll_utils"],
                    dict_objective=dict_objective,
                    additional_sampling_log=additional_sampling_log,
                    rng_state_array=rng_state_array,
                    rng_inc_array=rng_inc_array,
                )

            elif saver.check_need_to_update_memory(t):
                # print(f"updating memory at t={t}")
                additional_sampling_log["v"] = self.v.reshape((self.N, self.D)) * 1
                additional_sampling_log["type_t"] = type_t
                additional_sampling_log["accepted_t"] = accepted_t
                additional_sampling_log["log_proba_accept_t"] = log_proba_accept_t

                dict_objective, nll_full = posterior.compute_all_for_saver(
                    self.current["Theta"],
                    self.current["forward_map_evals"],
                    self.current["nll_utils"],
                )

                if t > T_BI:
                    dict_model_check = self._update_model_check_values(
                        dict_model_check,
                        posterior.likelihood,
                        nll_full,
                        dict_objective["objective"] * 1,
                    )

                rng_state_array, rng_inc_array = self.get_rng_state()

                saver.update_memory(
                    t,
                    Theta=self.current["Theta"],
                    forward_map_evals=self.current["forward_map_evals"],
                    nll_utils=self.current["nll_utils"],
                    dict_objective=dict_objective,
                    additional_sampling_log=additional_sampling_log,
                    rng_state_array=rng_state_array,
                    rng_inc_array=rng_inc_array,
                )

            else:
                pass

            if saver.check_need_to_save(t):
                # print(f"saving memory at t={t}")
                saver.save_to_file()

        # ---------
        dict_model_check = self._finalize_model_check_values(
            dict_model_check,
            likelihood=posterior.likelihood,
            forward_map_evals=self.current["forward_map_evals"],
            nll_full=nll_full,
        )

        saver.save_additional(
            list_arrays=[
                dict_model_check["clppd_online"],
                dict_model_check["p_values_y"],
                dict_model_check["p_values_llh"],
            ],
            list_names=["clppd", "p-values-y", "p-values-llh"],
        )
        return

    # def generate_new_sample_pmala_rmsprop(self, t, posterior):
    #     """generates a new sample using the position-dependent MALA transition kernel

    #     Parameters
    #     ----------
    #     t : int
    #         current iteration index
    #     score_model : ScoreModel
    #         negative log posterior class

    #     Returns
    #     -------
    #     accepted : bool
    #         wether or not the candidate was accepted
    #     log_proba_accept : float
    #         log of the acceptance proba
    #     """
    #     grad_t = self.current["grad"].flatten()

    #     # print(self.lambda_ + np.sqrt(self.v))
    #     diag_G_t = 1 / (self.lambda_ + np.sqrt(self.v))

    #     assert np.all(
    #         diag_G_t > 0
    #     ), f"{diag_G_t}, {self.lambda_ + np.sqrt(self.v)}, {self.v}"

    #     # generate random
    #     z_t = self.rng.standard_normal(size=self.N * self.D)
    #     z_t *= np.sqrt(self.eps0 * diag_G_t)

    #     # bias correction term
    #     if self.compute_correction_term:
    #         # recursive version
    #         # correction = -1 / 2 * diag_G_t ** 2 / np.sqrt(self.v) * self.u

    #         # only with corresponding term
    #         hess_diag_t = self.current["hess_diag"].flatten()
    #         correction = (
    #             -(1 - self.alpha)
    #             * self.alpha ** self.j_t
    #             * (diag_G_t ** 2)
    #             / np.sqrt(self.v)
    #             * grad_t
    #             * hess_diag_t
    #         )
    #         if np.sum(~np.isfinite(correction)) > 0:
    #             print(
    #                 f"num of nan in correction term: {np.sum(~np.isfinite(correction))}"
    #             )
    #         correction = np.nan_to_num(correction)  # ? nécessaire ?
    #     else:
    #         correction = np.zeros((self.N * self.D,))

    #     # combination
    #     mu_current = (
    #         self.current["Theta"].flatten()
    #         - self.eps0 / 2 * diag_G_t * grad_t
    #         + self.eps0 * correction
    #     )

    #     if self.stochastic:
    #         candidate = mu_current + z_t  # (N * D,)

    #         log_q_candidate_given_current = -1 / 2 * np.sum(np.log(diag_G_t)) - 1 / (
    #             2 * self.eps0
    #         ) * np.sum((candidate - mu_current) ** 2 / diag_G_t)

    #         # * compute log_q of candidate given current
    #         candidate_all = posterior.compute_all(
    #             candidate.reshape(self.N, self.D),
    #         )
    #         grad_cand = candidate_all["grad"].flatten()
    #         v_cand = self.alpha * self.v + (1 - self.alpha) * grad_cand ** 2
    #         diag_G_cand = 1 / (self.lambda_ + np.sqrt(v_cand))

    #         if self.compute_correction_term:
    #             hess_diag_cand = candidate_all["hess_diag"].flatten()

    #             correction_cand = -(
    #                 (1 - self.alpha)
    #                 * diag_G_cand ** 2
    #                 / np.sqrt(v_cand)
    #                 * grad_cand
    #                 * hess_diag_cand
    #             )
    #         else:
    #             correction_cand = np.zeros((self.N * self.D,))

    #         mu_cand = (
    #             candidate
    #             - self.eps0 / 2 * diag_G_cand * grad_cand
    #             + self.eps0 * correction_cand
    #         )

    #         log_q_current_given_candidate = -1 / 2 * np.sum(np.log(diag_G_cand)) - 1 / (
    #             2 * self.eps0
    #         ) * np.sum((self.current["Theta"].flatten() - mu_cand) ** 2 / diag_G_cand)

    #         # * compute proba accept
    #         logpdf_current = -self.current["objective"] * 1
    #         logpdf_candidate = -candidate_all["objective"] * 1

    #         log_proba_accept = (
    #             logpdf_candidate
    #             - logpdf_current
    #             + log_q_current_given_candidate
    #             - log_q_candidate_given_current
    #         )
    #         log_u = np.log(self.rng.uniform(0, 1))
    #         # print(
    #         #     f"{log_u:.2e}, {log_proba_accept:.4e}, {logpdf_candidate:.4e},, {logpdf_current:.4e}, {log_q_current_given_candidate:.4e}, {log_q_candidate_given_current:.4e}"
    #         # )

    #         if log_u < log_proba_accept:
    #             self.current = copy.copy(candidate_all)
    #             self.v = v_cand * 1
    #             assert np.sum(np.isnan(self.v)) == 0.0
    #             assert np.sum(np.isinf(self.v)) == 0.0

    #             self.j_t = np.zeros((self.N * self.D,))
    #             return True, log_proba_accept
    #         else:
    #             self.j_t += 1
    #             self.v = v_cand * 1
    #             assert np.sum(np.isnan(self.v)) == 0.0
    #             assert np.sum(np.isinf(self.v)) == 0.0
    #             return False, log_proba_accept

    #     # * in case we are doing optimization and not sampling
    #     candidate_all = posterior.compute_all(mu_current.reshape((self.N, self.D)))
    #     if candidate_all["objective"] < self.current["objective"]:
    #         self.current = copy.copy(candidate_all)
    #         accept = True
    #         proba = 1
    #     else:
    #         candidate = mu_current + z_t  # (N * D,)
    #         candidate_all = posterior.compute_all(candidate.reshape((self.N, self.D)))

    #         if candidate_all["objective"] < self.current["objective"]:
    #             self.current = copy.copy(candidate_all)
    #             accept = True
    #             proba = 1
    #         else:
    #             accept = False
    #             proba = 0

    #     grad_tp1 = candidate_all["grad"].flatten()
    #     self.v = self.alpha * self.v + (1 - self.alpha) * grad_tp1 ** 2

    #     assert np.sum(np.isnan(self.v)) == 0.0
    #     assert (
    #         np.sum(np.isinf(self.v)) == 0.0
    #     ), f"{candidate_all['Theta']}, {candidate_all['grad']}"
    #     assert np.sum(np.isnan(self.current["Theta"])) == 0.0

    #     return accept, proba

    def generate_new_sample_pmala_rmsprop(self, t: int, posterior: Posterior):
        """generates a new sample using the position-dependent MALA transition kernel

        Parameters
        ----------
        t : int
            current iteration index
        posterior : ScoreModel
            negative log posterior class

        Returns
        -------
        accepted : bool
            wether or not the candidate was accepted
        log_proba_accept : float
            log of the acceptance proba
        """
        if self.stochastic:
            accept_total = np.zeros((self.N,))
            log_proba_accept_total = np.zeros((self.N,))

            # * define proba of changing each pixel
            # * either uniformly or depending on their respective nll
            # if posterior.prior_spatial is not None:
            # n_sites = len(posterior.dict_sites)
            # idx_site = int(self.rng.integers(0, n_sites))
            list_idx = np.array(list(posterior.dict_sites.keys()))

            for idx_site in list_idx:
                idx_pix = posterior.dict_sites[idx_site]
                n_pix = idx_pix.size

                new_Theta = self.current["Theta"] * 1  # (N, D)
                grad_t = self.current["grad"][idx_pix, :] * 1
                v_current = self.v.reshape((self.N, self.D))[idx_pix, :] * 1

                # generate random
                diag_G_t = 1 / (self.lambda_ + np.sqrt(v_current))  # (n_pix, D)

                assert np.all(
                    diag_G_t > 0
                ), f"{diag_G_t}, {self.lambda_ + np.sqrt(self.v)}, {self.v}"

                z_t = self.rng.standard_normal(size=(n_pix, self.D))
                z_t *= np.sqrt(self.eps0 * diag_G_t)  # (n_pix, D)

                # bias correction term
                if self.compute_correction_term:
                    # recursive version
                    # correction = -1 / 2 * diag_G_t ** 2
                    # / np.sqrt(self.v) * self.u

                    # only with corresponding term
                    hess_diag_t = self.current["hess_diag"][idx_pix, :] * 1
                    j_t = self.j_t.reshape((self.N, self.D))[idx_pix, :] * 1

                    correction = (
                        -(1 - self.alpha)
                        * self.alpha**j_t
                        * (diag_G_t**2)
                        / np.sqrt(v_current)
                        * grad_t
                        * hess_diag_t
                    )  # (n_pix, D)
                    if np.sum(~np.isfinite(correction)) > 0:
                        n_inf = np.sum(~np.isfinite(correction))
                        print(f"num of nan in correction term: {n_inf}")
                    correction = np.nan_to_num(correction)  # ? nécessaire ?
                else:
                    correction = np.zeros((n_pix, self.D))

                # combination
                mu_current = (
                    new_Theta[idx_pix, :]
                    - self.eps0 / 2 * diag_G_t * grad_t
                    + self.eps0 * correction
                )  # (n_pix, D)

                candidate = mu_current + z_t  # (n_pix, D)

                log_q_candidate_given_current = -1 / 2 * np.sum(
                    np.log(diag_G_t), axis=1
                ) - 1 / (2 * self.eps0) * np.sum(
                    (candidate - mu_current) ** 2 / diag_G_t, axis=1
                )  # (n_pix,)

                shape_q = log_q_candidate_given_current.shape
                assert shape_q == (n_pix,), f"{shape_q}"

                # * compute log_q of candidate given current
                candidate_full = new_Theta * 1
                candidate_full[idx_pix, :] = mu_current * 1

                candidate_all = posterior.compute_all(
                    candidate_full,
                    compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
                )
                grad_cand = candidate_all["grad"][idx_pix, :] * 1
                v_cand = (
                    self.alpha * v_current + (1 - self.alpha) * grad_cand**2
                )  # (n_pix, D)
                diag_G_cand = 1 / (self.lambda_ + np.sqrt(v_cand))  # (n_pix, D)

                if self.compute_correction_term:
                    hess_diag_cand = candidate_all["hess_diag"][idx_pix, :] * 1

                    correction_cand = -(
                        (1 - self.alpha)
                        * diag_G_cand**2
                        / np.sqrt(v_cand)
                        * grad_cand
                        * hess_diag_cand
                    )
                else:
                    correction_cand = np.zeros((n_pix, self.D))

                mu_cand = (
                    candidate
                    - self.eps0 / 2 * diag_G_cand * grad_cand
                    + self.eps0 * correction_cand
                )  # (n_pix, D)

                log_q_current_given_candidate = -1 / 2 * np.sum(
                    np.log(diag_G_cand), axis=1
                ) - 1 / (2 * self.eps0) * np.sum(
                    (new_Theta[idx_pix, :] - mu_cand) ** 2 / diag_G_cand, axis=1
                )  # (n_pix,)

                shape_q = log_q_current_given_candidate.shape
                assert shape_q == (n_pix,), f"{shape_q}"

                # * compute proba accept
                logpdf_current = -self.current["objective_pix"][idx_pix]
                logpdf_candidate = -candidate_all["objective_pix"][idx_pix]

                shape_1 = logpdf_current.shape
                shape_2 = logpdf_candidate.shape
                assert shape_1 == (n_pix,), f"{shape_1}"
                assert shape_2 == (n_pix,), f"{shape_2}"

                log_proba_accept = (
                    logpdf_candidate
                    - logpdf_current
                    + log_q_current_given_candidate
                    - log_q_candidate_given_current
                )
                assert log_proba_accept.shape == (n_pix,)

                log_u = np.log(self.rng.uniform(0, 1, size=n_pix))
                accept_arr = log_u < log_proba_accept

                new_Theta[idx_pix, :] = np.where(
                    accept_arr[:, None] * np.ones((n_pix, self.D)),
                    candidate,  # (n_pix, D)
                    new_Theta[idx_pix, :],  # (n_pix, D)
                )

                accept_total[idx_pix] = accept_arr * 1
                log_proba_accept_total[idx_pix] = log_proba_accept * 1

                # update v and j
                v = self.v.reshape((self.N, self.D)) * 1
                v[idx_pix, :] = v_cand * 1
                self.v = v.flatten()

                j = self.j_t.reshape((self.N, self.D)) * 1
                j[idx_pix, :] = np.where(
                    accept_arr[:, None],
                    0.0,  # reset to 0 if accept
                    j[idx_pix, :] + 1,  # else add 1
                )
                self.j_t = j.flatten()

                if accept_arr.max() > 0:  # if at least one accept
                    self.current = posterior.compute_all(
                        new_Theta,
                        compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
                    )

            # after loop
            return accept_total.mean(), log_proba_accept_total.mean()

        else:  # if optimization
            grad_t = self.current["grad"].flatten()

            # print(self.lambda_ + np.sqrt(self.v))
            diag_G_t = 1 / (self.lambda_ + np.sqrt(self.v))

            assert np.all(
                diag_G_t > 0
            ), f"{diag_G_t}, {self.lambda_ + np.sqrt(self.v)}, {self.v}"

            # generate random
            z_t = self.rng.standard_normal(size=self.N * self.D)
            z_t *= np.sqrt(self.eps0 * diag_G_t)

            # combination
            mu_current = (
                self.current["Theta"].flatten() - self.eps0 / 2 * diag_G_t * grad_t
            )

            candidate_all = posterior.compute_all(
                mu_current.reshape((self.N, self.D)),
                compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
            )
            if candidate_all["objective"] < self.current["objective"]:
                self.current = copy.copy(candidate_all)
                accept = True
                proba = 1
            else:
                candidate = mu_current + z_t  # (N * D,)
                candidate_all = posterior.compute_all(
                    candidate.reshape((self.N, self.D)),
                    compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
                )

                if candidate_all["objective"] < self.current["objective"]:
                    self.current = copy.copy(candidate_all)
                    accept = True
                    proba = 1
                else:
                    accept = False
                    proba = 0

            grad_tp1 = candidate_all["grad"].flatten()
            self.v = self.alpha * self.v + (1 - self.alpha) * grad_tp1**2

            assert np.sum(np.isnan(self.v)) == 0.0
            assert (
                np.sum(np.isinf(self.v)) == 0.0
            ), f"{candidate_all['Theta']}, {candidate_all['grad']}"
            assert np.sum(np.isnan(self.current["Theta"])) == 0.0

            return accept, proba

    def generate_new_sample_mtm(
        self, t: int, posterior: Posterior  # , idx_site: Union[int, None] = None
    ):
        r"""generates a new sample using the MTM transition kernel

        Parameters
        ----------
        t : int
            current iteration index
        posterior : Posterior
            target posterior distribution to sample from

        Returns
        -------
        accepted : bool
            wether or not the candidate was accepted
        log_proba_accept : float
            log of the acceptance proba
        """
        new_Theta = self.current["Theta"] * 1  # (N, D)

        accept_total = np.zeros((self.N,))
        log_rg_total = np.zeros((self.N,))

        # * define proba of changing each pixel
        # * either uniformly or depending on their respective nll
        # if posterior.prior_spatial is not None:
        # n_sites = len(posterior.dict_sites)
        # idx_site = int(self.rng.integers(0, n_sites))
        list_idx = np.array(list(posterior.dict_sites.keys()))

        for idx_site in list_idx:
            idx_pix = posterior.dict_sites[idx_site]
            n_pix = idx_pix.size

            # * generate and evaluate candidates
            candidates_pix = np.zeros((n_pix, self.k_mtm + 1, self.D))
            candidates_pix[:, :-1, :] = self.generate_random_start_Theta_1pix(
                new_Theta, posterior, idx_pix
            )
            candidates_pix[:, -1, :] = self.current["Theta"][idx_pix, :] * 1
            candidates_pix = candidates_pix.reshape(
                (n_pix * (self.k_mtm + 1), self.D)
            )  # (n_pix * (k_mtm+1), D)

            neglogpdf_candidates = posterior.likelihood.neglog_pdf_candidates(
                candidates_pix,
                idx=idx_pix,
                Theta_t=new_Theta * 1,  # self.current["Theta"] * 1
            )  # (n_pix * (k_mtm+1),)
            assert neglogpdf_candidates.shape == (n_pix * (self.k_mtm + 1),)

            candidates_pix = candidates_pix.reshape((n_pix, self.k_mtm + 1, self.D))
            # assert np.allclose(candidates_pix[:, -1, :], self.current["Theta"][idx_pix, :]) -> validated

            neglogpdf_candidates = neglogpdf_candidates.reshape((n_pix, self.k_mtm + 1))

            # * if optimization: define challenger with conditional posterior
            # * instead of likelihood, and only keep if better than current
            if not self.stochastic:
                neglogpdf_candidates += posterior.partial_neglog_pdf_priors(
                    new_Theta.copy(), idx_pix, candidates_pix
                )  # (n_pix, k_mtm)
                idx_challengers = np.argmin(
                    neglogpdf_candidates[:, :-1], axis=1
                )  # (n_pix,)
                assert idx_challengers.shape == (n_pix,)

                neglogpdf_candidates_challengers = np.zeros((n_pix,))
                challengers = np.zeros((n_pix, self.D))
                for i in range(n_pix):
                    neglogpdf_candidates_challengers[i] = neglogpdf_candidates[
                        i, idx_challengers[i]
                    ]
                    challengers[i, :] = candidates_pix[i, idx_challengers[i], :]
                # neglogpdf_candidates_challengers = neglogpdf_candidates[
                #     np.arange(len(candidates_pix)), idx_challengers
                # ]
                assert neglogpdf_candidates_challengers.shape == (
                    n_pix,
                ), neglogpdf_candidates_challengers.shape

                # challengers = candidates_pix[
                #     np.arange(len(candidates_pix)), idx_challengers, :
                # ]
                assert challengers.shape == (n_pix, self.D), challengers.shape

                # * compute values of corresponding pixels in current x
                candidates_already_Theta = candidates_pix[:, -1, :] * 1
                neglogpdf_already_Theta = neglogpdf_candidates[:, -1] * 1
                assert candidates_already_Theta.shape == (n_pix, self.D)
                assert neglogpdf_already_Theta.shape == (n_pix,)

                # * select best pixels
                accept_arr = (
                    (neglogpdf_candidates_challengers < neglogpdf_already_Theta)
                    & np.isfinite(neglogpdf_candidates_challengers)
                    & np.isfinite(neglogpdf_already_Theta)
                )

                new_Theta[idx_pix, :] = np.where(
                    accept_arr[:, None] * np.ones((n_pix, self.D)),
                    challengers,  # (n_pix, D)
                    candidates_already_Theta,  # (n_pix, D)
                )

                # * save which pixels were accepted
                accept_total[idx_pix] = accept_arr * 1

            # *------
            # * if sampling
            else:
                if posterior.prior_spatial is not None:
                    nlratio_prior_proposal = utils.compute_nlratio_prior_proposal(
                        new_Theta * 1,
                        posterior.prior_spatial.list_edges,
                        posterior.prior_spatial.weights,
                        idx_pix,
                        candidates_pix,
                    )
                    shape_ = nlratio_prior_proposal.shape
                    assert shape_ == (n_pix, self.k_mtm + 1)
                    neglogpdf_candidates += nlratio_prior_proposal

                neglogpdf_candidates_min = np.amin(
                    neglogpdf_candidates, axis=1, keepdims=True
                )
                neglogpdf_candidates -= neglogpdf_candidates_min

                pdf_candidates = np.exp(-neglogpdf_candidates)  # (n_pix, k_mtm)

                log_numerators = np.log(np.sum(pdf_candidates[:, :-1], axis=1))
                # log_numerators = np.where(
                #     np.isinf(log_numerators), -1e15, log_numerators
                # )

                assert log_numerators.shape == (n_pix,), log_numerators.shape
                # assert np.sum(1 - np.isfinite(log_numerators)) == 0, log_numerators

                # * choose challenger candidate
                weights = softmax(-neglogpdf_candidates[:, :-1], axis=1)
                assert np.sum(1 - np.isfinite(weights)) == 0, weights

                idx_challengers = np.zeros((n_pix,), dtype=int)
                for i in range(n_pix):
                    idx_challengers[i] = self.rng.choice(
                        self.k_mtm,
                        p=weights[i],
                    )

                challengers = candidates_pix[
                    np.arange(n_pix), idx_challengers, :
                ]  # (n_pix, D)
                neglogpdf_challengers = neglogpdf_candidates[
                    np.arange(n_pix), idx_challengers
                ]

                shape_ = neglogpdf_challengers.shape
                assert shape_ == (n_pix,), shape_

                # * denominator
                log_denominators = np.log(
                    np.sum(pdf_candidates, axis=1) - np.exp(-neglogpdf_challengers)
                )
                # log_denominators = np.where(
                #     np.isinf(log_denominators), -1e15, log_denominators
                # )

                shape_ = log_denominators.shape
                assert shape_ == (n_pix,), shape_

                # assert np.sum(1 - np.isfinite(log_numerators)) == 0, log_numerators
                # assert np.sum(1 - np.isfinite(log_denominators)) == 0, log_denominators

                # * accept-reject
                log_rg = log_numerators - log_denominators
                log_rg = np.where(
                    np.isfinite(log_rg), log_rg, 1e-15
                )  # if either log_numerators or log_denominators is not finite, do not accept

                log_u = np.log(self.rng.uniform(0, 1, size=n_pix))
                accept_arr = log_u < log_rg

                new_Theta[idx_pix, :] = np.where(
                    accept_arr[:, None] * np.ones((n_pix, self.D)),
                    challengers,  # (n_pix, D)
                    candidates_pix[:, -1, :],  # (n_pix, D)
                )

                accept_total[idx_pix] = accept_arr * 1
                log_rg_total[idx_pix] = log_rg * 1

                # * re-initialize j for new point
                new_j_t = self.j_t.reshape((self.N, self.D))
                new_j_t[idx_pix, :] = np.where(
                    accept_arr[:, None], 0.0, new_j_t[idx_pix, :]
                )
                self.j_t = new_j_t.flatten()  # (ND,)

        # *------
        # * once all sites have been dealt with, update global parameters
        if accept_total.max() > 0:  # if at least one accept
            self.current = posterior.compute_all(
                new_Theta,
                compute_derivatives_2nd_order=self.compute_derivatives_2nd_order,
            )

            new_v = self.v.reshape((self.N, self.D))
            new_v = np.where(
                accept_total[:, None],
                self.alpha * new_v + (1 - self.alpha) * self.current["grad"] ** 2,
                new_v,
            )
            self.v = new_v.flatten()
            assert np.sum(np.isnan(self.v)) == 0.0
            assert np.sum(np.isinf(self.v)) == 0.0

        if not self.stochastic:
            return np.mean(accept_total), np.mean(accept_total)
        else:
            return np.mean(accept_total), np.mean(log_rg_total)



import copy
import os
from typing import Dict, List, Optional, Union

import numpy as np

from beetroots.inversion.plots.plots_estimator import PlotsEstimator
from beetroots.inversion.results.abstract_results import ResultsExtractor
from beetroots.inversion.results.utils.bayes_pval_plots import ResultsBayesPvalues
from beetroots.inversion.results.utils.clppd import ResultsCLPPD
from beetroots.inversion.results.utils.ess_plots import ResultsESS
from beetroots.inversion.results.utils.kernel import ResultsKernels
from beetroots.inversion.results.utils.lowest_obj_estimator import (
    ResultsLowestObjective,
)
from beetroots.inversion.results.utils.mc import ResultsMC
from beetroots.inversion.results.utils.mmse_ci import ResultsMMSEandCI
from beetroots.inversion.results.utils.objective import ResultsObjective
from beetroots.inversion.results.utils.regularization_weights import (
    ResultsRegularizationWeights,
)
from beetroots.inversion.results.utils.valid_mc import ResultsValidMC
from beetroots.inversion.results.utils.y_f_Theta import (
    ResultsDistributionComparisonYandFTheta,
)
from beetroots.modelling.posterior import Posterior
from beetroots.space_transform.abstract_transform import Scaler


class ResultsExtractorMCMC(ResultsExtractor):
    r"""extractor of inference results for the Markov chain data that was saved."""

    __slots__ = (
        "path_data_csv_out_mcmc",
        "path_img",
        "path_raw",
        "N_MCMC",
        "T_MC",
        "T_BI",
        "freq_save",
        "max_workers",
    )

    def __init__(
        self,
        path_data_csv_out_mcmc: str,
        path_img: str,
        path_raw: str,
        N_MCMC: int,
        T_MC: int,
        T_BI: int,
        freq_save: int,
        max_workers: int,
    ):
        r"""

        Parameters
        ----------
        path_data_csv_out_mcmc : str
            path to the csv file in which the performance of estimators is to be saved
        path_img : str
            path to the folder in which images are to be saved
        path_raw : str
            path to the raw ``.hdf5`` files
        N_MCMC : int
            number of Markov chains to run per posterior distribution
        T_MC : int
            total size of each Markov chain
        T_BI : int
            duration of the `Burn-in` phase
        freq_save : int
            frequency of saved iterates, 1 means that all iterates were saved (used to show correct Markov chain sizes in chain plots)
        max_workers : int
            maximum number of workers that can be used for results extraction
        """
        self.path_data_csv_out_mcmc = path_data_csv_out_mcmc
        r"""str: path to the csv file in which the performance of estimators is to be saved"""
        self.path_img = path_img
        r"""str: path to the folder in which images are to be saved"""
        self.path_raw = path_raw
        r"""str: path to the raw ``.hdf5`` files"""

        self.N_MCMC = N_MCMC
        r"""int: number of Markov chains to run per posterior distribution"""
        self.T_MC = T_MC
        r"""int: total size of each Markov chain"""
        self.T_BI = T_BI
        r"""int: duration of the `Burn-in` phase"""
        self.freq_save = freq_save
        r"""int: frequency of saved iterates, 1 means that all iterates were saved (used to show correct Markov chain sizes in chain plots)"""

        self.max_workers = max_workers
        r"""int: maximum number of workers that can be used for results extraction"""

    def main(
        self,
        posterior: Posterior,
        model_name: str,
        scaler: Scaler,
        list_names: List[str],
        list_idx_sampling: List[int],
        list_fixed_values: Union[List[float], np.ndarray],
        #
        plot_1D_chains: bool,
        plot_2D_chains: bool,
        plot_ESS: bool,
        plot_comparisons_yspace: bool,
        #
        estimator_plot: Optional[PlotsEstimator] = None,
        analyze_regularization_weight: bool = False,
        Theta_true_scaled: Optional[np.ndarray] = None,
        list_lines_fit: Optional[List[str]] = None,
        #
        list_lines_valid: List[str] = [],
        y_valid: Optional[np.ndarray] = None,
        sigma_a_valid: Optional[np.ndarray] = None,
        omega_valid: Optional[np.ndarray] = None,
        sigma_m_valid: Optional[np.ndarray] = None,
        point_challenger: Dict = {},
        list_CI: List[int] = [68, 90, 95, 99],
    ):
        r"""performs the data extraction, in this order:

        * step 1 : clppd (see ``beetroots.inversion.results.utils.clppd``)
        * step 2 : kernel analysis (see ``beetroots.inversion.results.utils.kernel``)
        * step 3 : objective evolution (see ``beetroots.inversion.results.utils.objective``) (the objective is the negative log posterior pdf)
        * step 4 : MAP estimator from samples (see ``beetroots.inversion.results.utils.lowest_obj_estimator``)
        * step 5 : deal with whole Markov chain for MMSE and histograms, in a pixel-wise approach to avoid overloading the memory (see ``beetroots.inversion.results.utils.mc``)
        * step 6 : save global MMSE performance (see ``beetroots.inversion.results.utils.mmse_ci``)
        * step 7 (if map) : plot maps of ESS (see ``beetroots.inversion.results.utils.ess_plots``)
        * step 8 : model checking with Bayesian p-value computation (see ``beetroots.inversion.results.utils.bayes_pval_plots``)
        * step 9 (if the true value is known): plot how many components have their true value between min and max of Markov chain (see ``beetroots.inversion.results.utils.valid_mc``)
        * step 10 : plot comparison of distributions of :math:`y_n` and :math:`f(\theta_n)` for all :math:`n \in [\![1, N]\!]` (see ``beetroots.inversion.results.utils.y_f_Theta``)
        * step 11 (if ``analyze_regularization_weight``) : analysis of the regularization weight :math:`\tau` automatic tuning (see ``beetroots.inversion.results.utils.regularization_weights``)

        Parameters
        ----------
        posterior : Posterior
            probability distribution used to generate the Markov chain(s)
        model_name : str
            name of the model, used to identify the posterior distribution
        scaler : Scaler
            contains the transformation of the Theta values from their natural space to their scaled space (in which the sampling happens) and its inverse
        list_names : List[str]
            names of the D physical parameters to appear in plots (for instance, $P_{th}$ for thermal pressure)
        list_idx_sampling : List[int]
            indices of the physical parameters that were sampled (the other ones were fixed)
        list_fixed_values : List[float]
            list of used values for the parameters fixed during the sampling
        plot_1D_chains : bool
            wether to plot each of the :math:`N \times D` 1D chains and histograms for each physical parameter :math:`\theta_{nd}`
        plot_2D_chains : bool
            wether to plot each of the :math:`N \times D \times (D-1)` 2D chains and histograms for pairs of parameters :math:`(\theta_{n d_1}, \theta_{n d_2})` with :math:`1 \leq d_1 < d_2 \leq D`
        plot_ESS : bool
            wether to plot the Effective sample size maps (only used when :math:`N > 1`)
        plot_comparisons_yspace : bool
            whether to plot comparisons of the distribution on :math:`y_n` and :math:`\mathcal{A}(f(\theta_n))` (with :math:`\mathcal{A}` the noise model). Offers a visualization to understand the model checking based on the Bayesian p-value :cite:p:`paludProblemesInversesTest2023a`
        estimator_plot : Optional[PlotsEstimator], optional
            object used to plot the estimator figures, by default None
        analyze_regularization_weight : bool, optional
            wether to analyze the evaluation of the regularization weight :math:`\tau`, by default False
        Theta_true_scaled : Optional[np.ndarray], optional
            true value for the inferred physical parameter :math:`\Theta` (only possible for toy cases), by default None
        list_lines_fit : Optional[List[str]], optional
            names of the observables used for the inversion, by default None
        list_lines_valid : List[str], optional
            names of the available observables not used for the inversion (can be used for model checking), by default []
        y_valid : Optional[np.ndarray], optional
            observation values for the observables not used for inversion. If provided, must have shape (N, L_valid). by default None
        sigma_a_valid : Optional[np.ndarray], optional
            additive noise standard deviation values for the observables not used for inversion. If provided, must have shape (N, L_valid). by default None
        omega_valid : Optional[np.ndarray], optional
            censor threshold values for the observables not used for inversion. If provided, must have shape (N, L_valid)., by default None
        sigma_m_valid : Optional[np.ndarray], optional
            multiplicative noise standard deviation values for the observables not used for inversion. If provided, must have shape (N, L_valid)., by default None
        point_challenger : Dict, optional
            other estimator that can come from the literature, provided to be compared with the inference results, by default {}
        list_CI : List[int], optional
            list of credibility intervals to evaluate (in percent), by default [68, 90, 95, 99]
        """
        list_mcmc_folders = [
            f"{x[0]}/mc_chains.hdf5"
            for x in os.walk(f"{self.path_raw}/{model_name}")
            if "mcmc_" in x[0]
        ]
        list_mcmc_folders.sort()

        chain_type = "mcmc"
        N = posterior.N * 1
        D_sampling = posterior.D * 1
        D = len(list_fixed_values)
        L = len(list_lines_fit)

        assert D >= D_sampling, f"should have D={D} >= D_sampling={D_sampling}"
        print(f"N = {N}, L (fit) = {L}, D_sampling = {D_sampling}, D = {D}")

        if estimator_plot is not None:
            map_shaper = copy.copy(estimator_plot.map_shaper)
        else:
            map_shaper = None

        list_fixed_values_scaled = list(
            posterior.likelihood.forward_map.arr_fixed_values
        )
        list_fixed_values_scaled = [v if v != 0 else None for v in list_fixed_values]

        # clppd
        ResultsCLPPD(
            model_name=model_name,
            chain_type=chain_type,
            path_img=self.path_img,
            path_data_csv_out=self.path_data_csv_out_mcmc,
            N_MCMC=self.N_MCMC,
            N=N,
            L=L,
        ).main(list_mcmc_folders, map_shaper)

        # kernel analysis
        ResultsKernels(
            model_name,
            chain_type,
            self.path_img * 1,
            self.N_MCMC * 1,
            self.T_MC * 1,
            self.freq_save * 1,
        ).main(list_mcmc_folders)

        # objective evolution
        if Theta_true_scaled is not None:
            forward_map_evals = posterior.likelihood.evaluate_all_forward_map(
                Theta_true_scaled,
                compute_derivatives=False,
                compute_derivatives_2nd_order=False,
            )
            nll_utils = posterior.likelihood.evaluate_all_nll_utils(
                forward_map_evals,
                compute_derivatives=False,
                compute_derivatives_2nd_order=False,
            )
            objective_true = posterior.neglog_pdf(
                Theta_true_scaled, forward_map_evals, nll_utils
            )

        else:
            objective_true = np.nan

        idx_lowest_obj, lowest_obj = ResultsObjective(
            model_name,
            chain_type,
            self.path_img * 1,
            self.N_MCMC * 1,
            self.T_MC * 1,
            self.T_BI * 1,
            self.freq_save * 1,
            N=N,
            D=D,
            L=L,
        ).main(list_mcmc_folders, objective_true)

        # MAP estimator from samples
        if Theta_true_scaled is not None:
            Theta_true_scaled_full = np.zeros((N, D))
            Theta_true_scaled_full[:, list_idx_sampling] += Theta_true_scaled

            for d in range(D):
                if list_fixed_values_scaled[d] is not None:
                    Theta_true_scaled_full[:, d] += list_fixed_values_scaled[d]
        else:
            Theta_true_scaled_full = None

        ResultsLowestObjective(
            model_name,
            chain_type,
            self.path_img,
            self.path_data_csv_out_mcmc,
            self.N_MCMC,
            self.T_MC,
            self.freq_save,
        ).main(
            list_mcmc_folders,
            lowest_obj,
            idx_lowest_obj,
            scaler,
            Theta_true_scaled_full,  # (N, D_sampling)
            list_idx_sampling,  # (D_sampling,)
            list_fixed_values,  # (D,)
            estimator_plot,
        )

        # deal with whole MC for MMSE and histograms
        if estimator_plot is not None:
            lower_bounds_lin = estimator_plot.lower_bounds_lin * 1
            upper_bounds_lin = estimator_plot.upper_bounds_lin * 1
        else:
            lower_bounds = posterior.prior_indicator.lower_bounds_full.reshape(
                (1, D),
            )
            upper_bounds = posterior.prior_indicator.upper_bounds_full.reshape(
                (1, D),
            )

            lower_bounds_lin = scaler.from_scaled_to_lin(lower_bounds)
            lower_bounds_lin = lower_bounds_lin.flatten()
            upper_bounds_lin = scaler.from_scaled_to_lin(upper_bounds)
            upper_bounds_lin = upper_bounds_lin.flatten()

        ResultsMC(
            model_name,
            chain_type,
            self.path_img,
            self.path_data_csv_out_mcmc,
            self.max_workers,
            self.N_MCMC,
            self.T_MC,
            self.T_BI,
            self.freq_save,
            N,
            list_idx_sampling,  # (D_sampling,)
            list_fixed_values,  # (D,)
            lower_bounds_lin,
            upper_bounds_lin,
            list_names,
        ).main(
            scaler,
            Theta_true_scaled_full,
            list_mcmc_folders,
            plot_ESS,
            plot_1D_chains,
            plot_2D_chains,
            plot_comparisons_yspace,
            point_challenger=point_challenger,
            list_CI=list_CI,
        )

        # save global MMSE performance
        # (to do now, once the MMSE if computed for all pixels)
        _ = ResultsMMSEandCI(
            model_name,
            self.path_img,
            self.path_data_csv_out_mcmc,
            N,
            D,
        ).main(
            posterior,
            scaler,
            estimator_plot,
            Theta_true_scaled_full,
            list_idx_sampling,
            list_fixed_values,
            list_CI,
        )

        # plot maps of ESS
        if plot_ESS and estimator_plot is not None:
            ResultsESS(
                model_name,
                self.path_img,
                self.path_data_csv_out_mcmc,
                N,
                D_sampling,
            ).main(
                map_shaper,
                list_names,
                list_idx_sampling,
            )

        ResultsBayesPvalues(
            model_name,
            chain_type,
            self.path_img,
            self.path_data_csv_out_mcmc,
            self.N_MCMC,
            N,
            D_sampling,
            plot_ESS,
        ).main(
            list_idx_sampling=list_idx_sampling,
            map_shaper=map_shaper if N > 1 else None,
        )

        # plot how many components have their true value
        # between min and max of MC
        if Theta_true_scaled is not None:
            ResultsValidMC(
                model_name=model_name,
                path_img=self.path_img,
                path_data_csv_out_mcmc=self.path_data_csv_out_mcmc,
                N_MCMC=self.N_MCMC,
                T_MC=self.T_MC,
                T_BI=self.T_BI,
                freq_save=self.freq_save,
                N=N,
                D_sampling=D_sampling,
            ).main(list_names, list_idx_sampling)

        # plot comparison of distributions of y and f(\theta)
        if plot_comparisons_yspace and list_lines_fit is not None:
            comparison_y_f_Theta = ResultsDistributionComparisonYandFTheta(
                model_name,
                self.path_img,
                self.path_data_csv_out_mcmc,
                self.N_MCMC,
                self.T_MC,
                self.T_BI,
                self.freq_save,
                N,
                D,
                list_idx_sampling,
                self.max_workers,
            )

            if hasattr(posterior.likelihood, "sigma_a"):
                sigma_a = posterior.likelihood.sigma_a
            elif hasattr(posterior.likelihood, "sigma"):
                sigma_a = posterior.likelihood.sigma
            else:
                sigma_a = np.zeros_like(posterior.likelihood.y)

            if hasattr(posterior.likelihood, "sigma_m"):
                sigma_m = posterior.likelihood.sigma_m
            else:
                sigma_m = np.zeros_like(posterior.likelihood.y)

            comparison_y_f_Theta.main(
                list_mcmc_folders,
                scaler,
                posterior.likelihood.forward_map,
                posterior.likelihood.y,
                posterior.likelihood.omega,
                sigma_a,
                sigma_m,
                list_lines_fit,
                "fit",
                point_challenger=point_challenger,
            )

            if len(list_lines_valid) > 0:
                comparison_y_f_Theta.main(
                    list_mcmc_folders,
                    scaler,
                    posterior.likelihood.forward_map,
                    y_valid,
                    omega_valid,
                    sigma_a_valid,
                    sigma_m_valid,
                    list_lines_valid,
                    "valid",
                    point_challenger=point_challenger,
                )
                posterior.likelihood.forward_map.restrict_to_output_subset(
                    list_lines_fit,
                )

        if analyze_regularization_weight:
            ResultsRegularizationWeights(
                model_name,
                self.path_img,
                self.path_data_csv_out_mcmc,
                self.N_MCMC,
                self.T_MC,
                self.T_BI,
                self.freq_save,
                D,
                list_names,
            ).main(list_mcmc_folders)

        print()
        return



"""(deprecated) Implementation of L2 norm on image gradient prior
"""

import numpy as np

from beetroots.modelling.priors.abstract_spatial_prior import SpatialPrior


class L22SpatialPrior(SpatialPrior):
    r"""L22 smooth spatial prior, valid for both 1D and 2D tensors. Its pdf is defined as

    .. math::

        \forall d \in [1, D], \quad p(x_d) \propto \exp \left[- \tau_d \Vert \nabla Theta_d \Vert_F^2 \right]

    where  :math:`\Vert \cdot \Vert_F` denotes the Fröbenius norm
    """

    def neglog_pdf(
        self,
        Theta: np.ndarray,
        with_weights: bool = True,
    ) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        neglog_p = np.zeros((self.D,))

        if self.list_edges.size > 0:
            neglog_p += np.sum(
                (Theta[self.list_edges[:, 1], :] - Theta[self.list_edges[:, 0], :])
                ** 2,
                axis=0,
            )
        if with_weights:
            neglog_p *= self.weights

        return neglog_p  # (D,)

    def neglog_pdf_one_pix(
        self,
        Theta: np.ndarray,
        n: int,
        list_pixel_candidates: np.ndarray,
    ) -> np.ndarray:
        """
        computes the neg log-prior when only one pixel is modified

        Parameters
        ----------
        Theta : np.ndarray of shape (N, D)
            current iterate

        n : int
            the index of the pixel to consider (0 <= n <= N - 1)

        list_pixel_candidates : np.ndarray of shape (N_candidates, D)
            the list of all candidates for pixel n

        Returns
        -------
        nl_priors : np.ndarray of shape (N_candidates,)
            the leg log-prior of the candidates
        """
        assert Theta.shape == (self.N, self.D)
        assert 0 <= n <= self.N - 1
        # TODO : à reprendre

        N_candidates = list_pixel_candidates.shape[0]

        neglog_p = np.zeros_like(list_pixel_candidates)  # (N_candidates, D)

        list_edges_pix = self.list_edges[(self.list_edges[:, 0] == n)]

        if list_edges_pix.size > 0:
            neglog_p += (
                Theta[list_edges_pix[:, 1], :] - Theta[list_edges_pix[:, 0], :]
            ) ** 2  # (N_candidates, D)
        neglog_p = neglog_p * self.weights[None, :]
        neglog_p = np.sum(neglog_p, axis=1)
        return neglog_p  # (D,)

    def gradient_neglog_pdf(self, x: np.ndarray) -> np.ndarray:
        assert Theta.shape == (self.N, self.D)

        g = np.zeros_like(Theta)
        for edge in self.list_edges:
            val = 2 * (Theta[edge[1]] - Theta[edge[0]])  # (D,)
            g[edge[0]] -= val
            g[edge[1]] += val

        g = self.weights[None, :] * g
        return g  # (N, D)

    def hess_diag_neglog_pdf(self, Theta: np.ndarray) -> np.ndarray:
        hess_diag = np.zeros_like(x, dtype=np.float64)
        if self.list_edges.size > 0:
            idx, counts = np.unique(self.list_edges.flatten(), return_counts=True)
            # print(counts.dtype, hess_diag.dtype)
            hess_diag[idx, :] += 2 * counts[:, None] * np.ones((idx.size, self.D))

        hess_diag = self.weights[None, :] * hess_diag
        return hess_diag  # (N, D)
